from neuralqa.reader import Reader


import tensorflow as tf
import numpy as np
import time
import logging

logger = logging.getLogger(__name__)


class BERTReader(Reader):
    def __init__(self, model_name, model_path, model_type="bert", **kwargs):
        Reader.__init__(self, model_name, model_path, model_type)
        # self.load_model(model_name, model_path, model_type)

    def get_best_start_end_position(self, start_scores, end_scores):
        answer_start = tf.argmax(start_scores, axis=1).numpy()[0]
        answer_end = (tf.argmax(end_scores, axis=1) + 1).numpy()[0]
        return answer_start, answer_end

    def get_chunk_answer_span(self, inputs):
        start_time = time.time()
        answer_start_scores, answer_end_scores = self.model(inputs)

        answer_start, answer_end = self.get_best_start_end_position(
            answer_start_scores, answer_end_scores)

        answer_end = answer_end - \
            1 if answer_end == answer_end_scores.shape[1] else answer_end

        answer_start_softmax_probability = tf.nn.softmax(
            answer_start_scores, axis=1).numpy()[0][answer_start]
        answer_end_softmax_probability = tf.nn.softmax(
            answer_end_scores, axis=1).numpy()[0][answer_end]

        answer = self.tokenizer.decode(
            inputs["input_ids"][0][answer_start:answer_end], skip_special_tokens=True)

        # if model predict first token 0 which is in the question as part of the answer, return nothing
        if answer_start == 0:
            answer = ""

        elapsed_time = time.time() - start_time
        return {"answer": answer, "took": elapsed_time,
                "start_probability": str(answer_start_softmax_probability),
                "end_probability": str(answer_end_softmax_probability),
                "probability": str(answer_end_softmax_probability + answer_start_softmax_probability / 2)
                }

    def token_chunker(self, question, context, max_chunk_size=512, stride=2, max_num_chunks=5):
        # we tokenize question and context once.
        # if question + context > max chunksize, we break it down into multiple chunks of question +
        # subsets of context with some stride overlap

        question_tokens = self.tokenizer.encode(question)
        context_tokens = self.tokenizer.encode(
            context, add_special_tokens=False)

        chunk_holder = []
        chunk_size = max_chunk_size - len(question_tokens) - 1
        # -1 for the 102 end token we append later
        current_pos = 0
        chunk_count = 0
        while current_pos < len(context_tokens) and current_pos >= 0:

            # we want to cap the number of chunks we create 
            if max_num_chunks and chunk_count >= max_num_chunks:
                break

            end_point = current_pos + \
                chunk_size if (current_pos + chunk_size) < len(context_tokens) - \
                1 else len(context_tokens) - 1
            token_chunk = question_tokens + \
                context_tokens[current_pos: end_point] + [102]

            # question type is 0, context type is 1, convert to tf
            token_type_ids = [0]*len(question_tokens) + \
                [1] * (len(token_chunk) - len(question_tokens))
            token_type_ids = tf.constant(
                token_type_ids, dtype='int32', shape=(1, len(token_type_ids)))

            # attend to every token
            attention_mask = tf.ones(
                (1, len(token_chunk)),  dtype=tf.dtypes.int32)

            # convert token chunk to tf
            token_chunk = tf.constant(
                token_chunk, dtype='int32', shape=(1, len(token_chunk)))

            chunk_holder.append(
                {"token_ids": token_chunk,
                 "context": self.tokenizer.decode(context_tokens[current_pos: end_point], skip_special_tokens=True),
                 "attention_mask":  attention_mask,
                 "token_type_ids": token_type_ids
                 })
            current_pos = current_pos + chunk_size - stride + 1
            chunk_count +=1 
            
        return chunk_holder

    def answer_question(self, question, context, max_chunk_size=512, stride=70):

        # chunk tokens
        chunked_tokens = self.token_chunker(
            question, context, max_chunk_size, stride)
        answer_holder = []
        for chunk in chunked_tokens:
            model_input = {"input_ids": chunk["token_ids"], "attention_mask":
                           chunk["attention_mask"], "token_type_ids": chunk["token_type_ids"]}
            answer = self.get_chunk_answer_span(model_input)
            if len(answer["answer"]) > 2:
                answer["question"] = question
                answer["context"] = chunk["context"].replace("##", "").replace(
                    answer["answer"], " <em>" + answer["answer"] + "</em> ")
                answer_holder.append(answer)
        return answer_holder

    def get_correct_span_mask(self, correct_index, token_size):
        span_mask = np.zeros((1, token_size))
        span_mask[0, correct_index] = 1
        span_mask = tf.constant(span_mask, dtype='float32')
        return span_mask

    def get_embedding_matrix(self):
        if "DistilBert" in type(self.model).__name__:
            return self.model.distilbert.embeddings.word_embeddings
        else:
            return self.model.bert.embeddings.word_embeddings

    # move this to some utils file
    def clean_tokens(self, gradients, tokens, token_types):
        """
        Clean the tokens and  gradients
        Remove "[CLS]","[CLR]", "[SEP]" tokens
        Reduce (mean) gradients values for tokens that are split ##
        """
        token_holder = []
        token_type_holder = []
        gradient_holder = []
        i = 0
        while i < len(tokens):
            if (tokens[i] not in ["[CLS]", "[CLR]", "[SEP]"]):
                token = tokens[i]
                conn = gradients[i]
                token_type = token_types[i]
                if i < len(tokens)-1:
                    if tokens[i+1][0:2] == "##":
                        token = tokens[i]
                        conn = gradients[i]
                        j = 1
                        while i < len(tokens)-1 and tokens[i+1][0:2] == "##":
                            i += 1
                            token += tokens[i][2:]
                            conn += gradients[i]
                            j += 1
                        conn = conn / j
                token_holder.append(token)
                token_type_holder.append(token_type)
                gradient_holder.append(conn)
            i += 1
        return gradient_holder, token_holder, token_type_holder

    def get_gradient(self, question, context):
        """Return gradient of input (question) wrt to model output span prediction

        Args:
            question (str): text of input question
            context (str): text of question context/passage
            model (QA model): Hugging Face BERT model for QA transformers.modeling_tf_distilbert.TFDistilBertForQuestionAnswering, transformers.modeling_tf_bert.TFBertForQuestionAnswering
            tokenizer (tokenizer): transformers.tokenization_bert.BertTokenizerFast 

        Returns:
            (tuple): (gradients, token_words, token_types, answer_text)
        """

        embedding_matrix = self.get_embedding_matrix()

        encoded_tokens = self.tokenizer.encode_plus(
            question, context, add_special_tokens=True, return_token_type_ids=True, return_tensors="tf")
        token_ids = list(encoded_tokens["input_ids"].numpy()[0])
        vocab_size = embedding_matrix.get_shape()[0]

        # convert token ids to one hot. We can't differentiate wrt to int token ids hence the need for one hot representation
        token_ids_tensor = tf.constant([token_ids], dtype='int32')
        token_ids_tensor_one_hot = tf.one_hot(token_ids_tensor, vocab_size)

        with tf.GradientTape(watch_accessed_variables=False) as tape:
            # (i) watch input variable
            tape.watch(token_ids_tensor_one_hot)

            # multiply input model embedding matrix; allows us do backprop wrt one hot input
            inputs_embeds = tf.matmul(
                token_ids_tensor_one_hot, embedding_matrix)

            # (ii) get prediction
            start_scores, end_scores = self.model(
                {"inputs_embeds": inputs_embeds, "token_type_ids": encoded_tokens["token_type_ids"], "attention_mask": encoded_tokens["attention_mask"]})
            answer_start, answer_end = self.get_best_start_end_position(
                start_scores, end_scores)

            start_output_mask = self.get_correct_span_mask(
                answer_start, len(token_ids))
            end_output_mask = self.get_correct_span_mask(
                answer_end, len(token_ids))

            # zero out all predictions outside of the correct span positions; we want to get gradients wrt to just these positions
            predict_correct_start_token = tf.reduce_sum(
                start_scores * start_output_mask)
            predict_correct_end_token = tf.reduce_sum(
                end_scores * end_output_mask)

            # (iii) get gradient of input with respect to both start and end output
            gradient_non_normalized = tf.norm(
                tape.gradient([predict_correct_start_token, predict_correct_end_token], token_ids_tensor_one_hot), axis=2)

            # (iv) normalize gradient scores and return them as "explanations"
            gradient_tensor = (
                gradient_non_normalized /
                tf.reduce_max(gradient_non_normalized)
            )
            gradients = gradient_tensor[0].numpy().tolist()

            token_words = self.tokenizer.convert_ids_to_tokens(token_ids)
            token_types = list(
                encoded_tokens["token_type_ids"].numpy()[0].tolist())
            answer_text = self.tokenizer.decode(
                token_ids[answer_start:answer_end],  skip_special_tokens=True)

            # clean up gradients and words
            gradients, token_words, token_types = self.clean_tokens(
                gradients, token_words, token_types)
            return gradients, token_words, token_types, answer_text

    def explain_model(self, question, context, explain_method="gradient"):
        if explain_method == "gradient":
            return self.get_gradient(question, context)



"""
Definitions of click options shared by several CLI commands.
"""
import click


HOST = click.option("--host", "-h", default="127.0.0.1",
                    help="The network address to listen on (default: 127.0.0.1). "
                         "Use 0.0.0.0 to bind to all addresses if you want to access the tracking "
                         "server from other machines.")

PORT = click.option("--port", "-p", default=5000,
                    help="The port to listen on (default: 5000).")


WORKERS = click.option("--workers", "-w", default=1,
                       help="Number of uviicorn worker processes to handle requests (default: 1).")

MAX_DOCS = click.option("--max-docs", "-md", default=2000,
                       help="Maximum number of sample documents to import when loading sample data into local index")                       

CONFIG_PATH = click.option("--config-path", "-cp", default=None,
                           help="Path to a yaml file containing config for neuralqa. "
                           "If none is provided, the default config.yaml is copied to the current directory.")



from elasticsearch import Elasticsearch
import os
import zipfile
import shutil
import urllib.request
import logging
import lzma
import json
import tarfile
import hashlib


logger = logging.getLogger(__name__)
# index settings with analyzer to automatically remove stop words
index_settings = {
    "settings": {
        "analysis": {
            "analyzer": {
                "stop_analyzer": {
                    "type": "standard",
                    "stopwords": "_english_"
                }
            }
        }
    },
    "mappings": {
        "properties": {
            "casebody": {
                "type": "text",
                "analyzer": "stop_analyzer"
            },
            "name": {
                "type": "text",
                "analyzer": "stop_analyzer"
            }
        }
    }
}


def create_index_from_json(index_name, file_path, max_docs=None):
    """Create an index from json file formats.
    Read each file line by line, parse each line as json
        jsonl.xz
        json  : must be a json file containing a list

    Arguments:
        file_path {str} -- path to case.law bulk file
    Keyword Arguments:
        max_docs {int} -- maximum size of records to use in creating index.
        small default can be used to enable quick testing (e.g: {2000}).
        set this to None to use the entire data file.
    """

    es = Elasticsearch([{'host': 'localhost', 'port': 9200}])

    es.indices.create(
        index=index_name, body=index_settings, ignore=400)

    extension = os.path.splitext(file_path)[1]
    logger.info(">> Creating index using file " + file_path)
    i = 0
    if extension == ".xz":
        with lzma.open(file_path) as f:
            for line in f:
                i += 1
                line = json.loads(str(line, 'utf8'))
                try:
                    index_status = es.index(
                        index=index_name, id=i, body=line)
                except Exception as e:
                    logger.info(
                        "An error has occurred while creating index " + str(e))
                    break
                # logger.info(index_status)
                if (i > max_docs):
                    break
    logger.info(">> Creating index complete ")
    os.remove(file_path)


def import_scotus_files(max_docs=2000):
    scotus_url = "https://www.courtlistener.com/api/bulk-data/opinions/scotus.tar.gz"
    scotus_dir = "scotusdata"
    index_name = "supremecourt"

    if (not os.path.exists(scotus_dir)):
        os.makedirs(scotus_dir, exist_ok=True)
        logger.info(">>> Downloading supreme court case data")
        ftpstream = urllib.request.urlopen(scotus_url)
        thetarfile = tarfile.open(fileobj=ftpstream, mode="r|gz")
        thetarfile.extractall(path=scotus_dir)
        logger.info(">>> Download completed ")

    logger.info(">> Creating %s index using %s documents",
                index_name, str(max_docs))
    scotus_files = os.listdir(scotus_dir)

    es = Elasticsearch([{'host': 'localhost', 'port': 9200}])

    es.indices.create(
        index=index_name, body=index_settings, ignore=400)

    i = 0
    for file_path in (scotus_files):
        with open("scotusdata/" + file_path) as json_file:
            scotus_case = json.load(json_file)
            case = {"author": scotus_case["author"],
                    "casebody": scotus_case["plain_text"]}
            if (scotus_case["plain_text"] != ""):
                try:
                    index_status = es.index(
                        index=index_name, id=scotus_case["id"], body=case)
                except Exception as e:
                    logger.info(
                        "An error has occurred while creating index " + str(e))
                    break
                i += 1
        if (i > max_docs):
            break

    logger.info(">> Index creation complete.")


def download_data(data_url, source_name):
    """Download Zip datafile from case.law
    Arguments:
        data_url {str} -- url path dataset
        source_name {str} -- name for dataset
    """
    # create data directory
    os.makedirs("data", exist_ok=True)
    # download data from caselaw
    zip_file_path = source_name + ".zip"
    logger.info("Downloading data file for " + source_name)
    urllib.request.urlretrieve(data_url, zip_file_path)
    logger.info(">> Downloaded data file " + zip_file_path)

    extract_dir = "temp" + source_name
    with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
        zip_ref.extractall(extract_dir)
    data_file = os.path.join(extract_dir, os.listdir(
        extract_dir)[0], "data", "data.jsonl.xz")
    final_file_path = os.path.join("data", source_name + "jsonl.xz")
    shutil.copyfile(data_file, final_file_path)
    logger.info(">> Extracted and moved jsonl file to data folder")
    shutil.rmtree(extract_dir)
    os.remove(zip_file_path)
    return final_file_path


def import_sample_data(max_docs=2000):
    """This method downloads several datasets and builds an 
    elasticsearch index using the downloaded data. 
    Caselaw
    Supreme Court Cases
    Medical data

    Args:
        max_docs (int, optional): [description]. Defaults to 2000.
    """
    caselaw_data_paths = [
        ["https://api.case.law/v1/bulk/22411/download/", "newmexico"]
    ]
    for data_path in caselaw_data_paths:
        file_path = download_data(data_path[0], data_path[1])
        create_index_from_json("cases", file_path, max_docs=max_docs)

    # import_scotus_files(max_docs=max_docs)
    # import_medical_data(max_docs=max_docs)


def parse_field_content(field_name, content):
    """Parse content fields if nested using dot notation, else return content as is.
    e.g. for acrray content and field_name casebody.data.opinions.text, we return
    content[casebody][data][opinions][text]. If any nest level is an array we return only the 
    first instance of this array. e.g. if opinions is an array, we return  
    content[casebody][data][opinions][0][text]. 

    Args:
        field_name ([str]): [description]
        content ([dict]): [description]

    Returns:
        [str]: content of field
    """

    if ("." not in field_name):
        return content[field_name]
    else:
        fields = field_name.split(".")
        for field in fields:
            content = content[field]
            if (isinstance(content, list)):
                content = content[0]
        return content




from neuralqa.reader import BERTReader
import logging

logger = logging.getLogger(__name__)


class ReaderPool():
    def __init__(self, models):
        self._selected_model = models["selected"]
        self.reader_pool = {}
        for model in models["options"]:
            if (model["type"] == "bert" or model["type"] == "distilbert"):
                self.reader_pool[model["value"]] = BERTReader(
                    model["name"], model["value"])

    @property
    def model(self):
        return self.reader_pool[self.selected_model]

    @property
    def selected_model(self):
        return self._selected_model

    @selected_model.setter
    def selected_model(self, selected_model):

        if (selected_model in self.reader_pool):
            self._selected_model = selected_model
        else:
            if (len(self.reader_pool) > 0):
                default_model = next(iter(self.reader_pool))
                logger.info(
                    ">> Model you are attempting to use %s does not exist in model pool. Using the following default model instead %s ", selected_model, default_model)
                self._selected_model = default_model
            else:
                logger.info(
                    ">> No reader has been specified in config.yaml.")
                self._selected_model = None



from neuralqa.retriever import Retriever
from neuralqa.utils import parse_field_content
from elasticsearch import Elasticsearch, ConnectionError, NotFoundError
import logging


logger = logging.getLogger(__name__)


class ElasticSearchRetriever(Retriever):
    def __init__(self, index_type="elasticsearch", host="localhost", port=9200, **kwargs):
        Retriever.__init__(self, index_type)

        self.username = ""
        self.password = ""
        self.body_field = ""
        self.host = host
        self.port = port

        allowed_keys = list(self.__dict__.keys())
        self.__dict__.update((k, v)
                             for k, v in kwargs.items() if k in allowed_keys)
        self.es = Elasticsearch([{'host': self.host, 'port': self.port}])
        self.isAvailable = self.es.ping()

        rejected_keys = set(kwargs.keys()) - set(allowed_keys)

        if rejected_keys:
            raise ValueError(
                "Invalid arguments in ElasticSearchRetriever constructor:{}".format(rejected_keys))

    def run_query(self, index_name, search_query, max_documents=5, fragment_size=100, relsnip=True, num_fragments=5, highlight_tags=True):

        tags = {"pre_tags": [""], "post_tags": [
            ""]} if not highlight_tags else {}
        highlight_params = {
            "fragment_size": fragment_size,
            "fields": {
                self.body_field: tags
            },
            "number_of_fragments": num_fragments
        }

        search_query = {
            "_source": {"includes": [self.body_field]},
            "query": {
                "multi_match": {
                    "query":    search_query,
                    "fields": [self.body_field]
                }
            },
            "size": max_documents
        }

        status = True
        results = {}

        if (relsnip):
            # search_query["_source"] = {"includes": [""]}
            search_query["highlight"] = highlight_params
        # else:
        #     search_query["_source"] = {"includes": [self.body_field]}

        try:
            query_result = self.es.search(
                index=index_name, body=search_query)

            # RelSnip: for each document, we concatenate all
            # fragments in each document and return as the document.
            highlights = [" ".join(hit["highlight"][self.body_field])
                          for hit in query_result["hits"]["hits"] if "highlight" in hit]
            docs = [parse_field_content(self.body_field, hit["_source"])
                    for hit in query_result["hits"]["hits"] if "_source" in hit]
            took = query_result["took"]
            results = {"took": took,  "highlights": highlights, "docs": docs}

        except (ConnectionRefusedError, NotFoundError, Exception) as e:
            status = False
            results["errormsg"] = str(e)

        results["status"] = status
        return results

    def test_connection(self):
        try:
            self.es.cluster.health()
            return True
        except ConnectionError:
            return False
        except Exception as e:
            logger.info(
                'An unknown error occured connecting to ElasticSearch: %s' % e)
            return False



# # temporary file for testing search abstractions
# # TODO: Delete this file!!

# from elasticsearch import Elasticsearch, ConnectionError

# es = Elasticsearch([{'host': "localhost", 'port': 9200}])

# def parse_field_content(field_name, content):
#     """Parse content fields if nested using dot notation, else return content as is.
#     e.g. for acrray content and field_name casebody.data.opinions.text, we return
#     content[casebody][data][opinions][text]. If any nest level is an array we return only the
#     first instance of this array. e.g. if opinions is an array, we return
#     content[casebody][data][opinions][0][text].

#     Args:
#         field_name ([str]): [description]
#         content ([dict]): [description]

#     Returns:
#         [str]: content of field
#     """

#     if ("." not in field_name):
#         return content[field_name]
#     else:
#         fields = field_name.split(".")
#         for field in fields:
#             print("\t",content.keys(), field)
#             content =  content[field]
#             if (isinstance(content, list)):
#                 content = content[0]
#         return content

# def run_query(index_name, search_query, body_field, secondary_fields, max_documents=5, highlight_span=100, relsnip=True, num_fragments=5):

#     tags = {"pre_tags": [""], "post_tags": [""]}
#     highlight_params = {
#         "fragment_size": highlight_span,
#         "fields": {
#             body_field: tags
#         },
#         "number_of_fragments": num_fragments,
#         # "order": "score"
#     }

#     search_query = {
#         "query": {
#             "multi_match": {
#                 "query":    search_query,
#                 "fields": [body_field]
#             }
#         },
#         "_source": {"includes": [body_field]},
#         "size": max_documents
#     }

#     status = True
#     results = {}

#     if (relsnip):
#         search_query["highlight"] = highlight_params
#     # else:
#     #     search_query["_source"] = {"includes": [body_field]}

#     try:
#         # print(search_query)
#         query_result = es.search(index=index_name, body=search_query)
#         # RelSnip: for each document, we concatenate all
#         # fragments in each document and return as the document.

#         highlights = [" *** ".join(hit["highlight"][body_field])
#                         for hit in query_result["hits"]["hits"] if "highlight" in hit]
#         docs = [parse_field_content(body_field, hit["_source"])
#                 for hit in query_result["hits"]["hits"] if hit["_source"]]
#         took = query_result["took"]
#         results = {"took": took,  "highlight": highlights, "docs": docs}

#     except (ConnectionRefusedError, Exception) as e:
#         status = False
#         results["errormsg"] = str(e)

#     results["status"] = status
#     return results


# search_query = "what arson"
# max_documents = 2
# highlight_span = 50
# body_field = "casebody.data.opinions.text"
# secondary_fields = ["author"]
# num_fragments = 2

# results = run_query("cases", search_query, body_field, secondary_fields,
#                     max_documents=5, highlight_span=highlight_span, relsnip=True, num_fragments=num_fragments)
# print(results)
import logging
from neuralqa.expander import MLMExpander

logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)
# logging.getLogger("root").setLevel(logging.INFO)

expander = MLMExpander()
expanded_query = expander.expand_query(
    "what is the goal of the fourth amendment?  ")
terms = " ".join([term["token"] for term in expanded_query["terms"]])
print(expanded_query)




from neuralqa.expander import MLMExpander
import logging

logger = logging.getLogger(__name__)


class ExpanderPool():
    def __init__(self, expanders):
        self._selected_expander = expanders["selected"]
        self.expander_pool = {}
        for expander in expanders["options"]:
            if (expander["type"] == "maskedlm"):
                self.expander_pool[expander["value"]] = MLMExpander(
                    model_path=expander["value"])

    @property
    def expander(self):
        return self.expander_pool[self.selected_expander]

    @property
    def selected_expander(self):
        return self._selected_expander

    @selected_expander.setter
    def selected_expander(self, selected_expander):
        if (selected_expander in self.expander_pool):
            self._selected_expander = selected_expander
        else:
            if (len(self.expander_pool) > 0):
                default_expander = next(iter(self.expander_pool))
                logger.info(
                    ">> Expander you are attempting to use %s does not exist in expander pool. Using the following default expander instead %s ", selected_expander, default_expander)
                self._selected_expander = default_expander
            else:
                logger.info(
                    ">> No expander has been specified in config.yaml.")
                self._selected_expander = None



import yaml
import os
import logging
import shutil


logger = logging.getLogger(__name__)


class ConfigParser:
    def __init__(self, config_path):

        module_file_path = os.path.dirname(os.path.abspath(__file__))
        self.default_config_path = os.path.join(
            module_file_path, "../config_default.yaml")
        self.current_config_path = os.path.join(os.getcwd(), "config.yaml")

        if config_path and os.path.exists(config_path):
            self.config = self.load_config(config_path)
            # else:
            #     logger.info("Supplied config file does not exist. " +
            #                  os.path.join(os.getcwd(), config_path))
            #     logger.info("Creating new config file at " +
            #                  self.current_config_path)
            #     self.config = self.load_default_config()
        else:

            if (config_path and not os.path.exists(config_path)):
                logger.info(">> Supplied config file does not exist. " +
                            os.path.join(os.getcwd(), config_path))

            if os.path.exists(self.current_config_path):
                logger.info(">> Found config.yaml file found in current directory " +
                            self.current_config_path)
                self.config = self.load_config(self.current_config_path)
            else:
                logger.info("vCreating new config file at " +
                            self.current_config_path)
                shutil.copyfile(self.default_config_path,
                                self.current_config_path)
                self.config = self.load_default_config()

    def load_default_config(self):
        with open(self.default_config_path) as f:
            default_config = yaml.safe_load(f)
        return default_config

    def load_config(self, config_path):
        """Specially load a config file path. 
        Will first load the default config file, and update its values with 
        the content of the file in config_path.

        Args:
            config_path ([type]): [description]

        Returns:
            [type]: [description]
        """
        default_config = self.load_default_config()

        with open(config_path) as f:
            config = yaml.safe_load(f)

        default_config.update(config)
        return default_config



from neuralqa.expander import Expander
import logging
from transformers import AutoTokenizer, TFBertForMaskedLM
import tensorflow as tf
import time
import spacy
logger = logging.getLogger(__name__)


class MLMExpander(Expander):
    def __init__(self, index_type="mlm", model_path="bert-base-uncased", **kwargs):
        Expander.__init__(self, index_type)

        self.candidate_pos = ["NOUN", "ADJ", "ADV"]
        self.model_path = model_path

        allowed_keys = list(self.__dict__.keys())
        self.__dict__.update((k, v)
                             for k, v in kwargs.items() if k in allowed_keys)
        rejected_keys = set(kwargs.keys()) - set(allowed_keys)
        if rejected_keys:
            raise ValueError(
                "Invalid arguments in ElasticSearchRetriever constructor:{}".format(rejected_keys))

        logger.info(
            ">> loading HF model for Query Expansion from " + model_path)
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_path, use_fast=True)
        self.model = TFBertForMaskedLM.from_pretrained(
            self.model_path, from_pt=True)
        logger.info(">> Loading Spacy NLP model ")

        try:
            self.nlp = spacy.load('en_core_web_md')
        except OSError:
            logger.info(
                "Downloading language model for the spaCy POS tagger (don't worry, this will only happen once)")
            from spacy.cli import download
            download('en_core_web_md')
            self.nlp = spacy.load('en_core_web_md')
            # self.nlp = en_core_web_md.load()
        # logger.info(">> Spacy nlp model loaded ")

    def predict_mask(self, sequence, model, tokenizer, top_n=2):
        input = tokenizer.encode(sequence, return_tensors="tf")
        mask_token_index = tf.where(input == tokenizer.mask_token_id)[0, 1]
        token_logits = model(input)[0]
        mask_token_logits = token_logits[0, mask_token_index, :]

        probabilities = tf.nn.softmax(mask_token_logits)
        topk = tf.math.top_k(probabilities, top_n)
        top_n_probs, top_n_tokens = topk.values.numpy(), topk.indices.numpy()
        results = [{"token": tokenizer.decode([top_n_tokens[i]]), "probability": float(top_n_probs[i])}
                   for i in range(len(top_n_probs))]
        # print(results)
        return results

    def expand_query(self, query, top_n=3, threshold=0.2):
        start_time = time.time()

        doc = self.nlp(query)
        query_tokens = [str(token) for token in doc]
        new_terms = []
        candidate_expansions = []
        for i, token in enumerate(doc):
            if (token.pos_ in self.candidate_pos):
                temp_doc = query_tokens.copy()
                temp_doc[i] = self.tokenizer.mask_token
                temp_doc = " ".join(temp_doc)
                pred_tokens = self.predict_mask(
                    temp_doc, self.model, self.tokenizer, top_n=top_n)
                candidate_expansions.append(
                    {"token": str(token), "expansion": pred_tokens, "token_index": i})
                new_terms = new_terms + pred_tokens

        elapsed_time = time.time() - start_time

        terms_list = []
        seen_terms = []
        # remove punctuation,  low probability, words subwords, duplicates
        for token in new_terms:
            if token["token"].isalnum() and token["probability"] > threshold and "#" not in token["token"] and token["token"] not in query and token["token"] not in seen_terms:
                terms_list.append(token)
                seen_terms.append(token["token"])

        result = {
            "terms": terms_list,
            "query": query_tokens,
            "expansions": candidate_expansions,
            "took": elapsed_time
        }
        return result





import tensorflow as tf
import numpy as np
from transformers import AutoTokenizer, TFAutoModelForQuestionAnswering
import time
import logging


logger = logging.getLogger(__name__)


class Reader:
    def __init__(self, model_name, model_path, model_type, **kwargs):
        self.load_model(model_name, model_path, model_type)

    def load_model(self, model_name, model_path, model_type):
        logger.info(">> Loading HF model " +
                    model_name + " from " + model_path)
        self.type = model_type
        self.name = model_name
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_path, use_fast=True)
        self.model = TFAutoModelForQuestionAnswering.from_pretrained(
            model_path, from_pt=True)





from neuralqa.reader import BERTReader, ReaderPool
from neuralqa.server.routehandlers import Handler
from neuralqa.retriever import ElasticSearchRetriever, RetrieverPool
from neuralqa.utils import ConfigParser
from neuralqa.expander import ExpanderPool

import os
import logging
import time
import uvicorn
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
# from fastapi.middleware.cors import CORSMiddleware


logger = logging.getLogger(__name__)

config_path = os.environ.get("NEURALQA_CONFIG_PATH")
app_config = ConfigParser(config_path)

app = FastAPI()
api = FastAPI(root_path="/api")


# origins = [
#     "http://localhost",
#     "http://localhost:3000",
# ]

# app.add_middleware(
#     CORSMiddleware,
#     allow_origins=origins,
#     allow_credentials=True,
#     allow_methods=["*"],
#     allow_headers=["*"],
# )
root_file_path = os.path.dirname(os.path.abspath(__file__))
static_folder_root = os.path.join(root_file_path, "ui/build")

app.mount("/api", api)
app.mount("/", StaticFiles(directory=static_folder_root, html=True),
          name="ui")


@api.get('/config')
async def ui_config():
    config = app_config.config["ui"]
    # show only listed models to ui
    config["queryview"]["options"]["relsnip"] = app_config.config["relsnip"]
    config["queryview"]["options"]["samples"] = app_config.config["samples"]
    config["queryview"]["options"]["expander"] = app_config.config["expander"]
    config["queryview"]["options"]["reader"] = app_config.config["reader"]
    config["queryview"]["options"]["retriever"] = app_config.config["retriever"]
    return config

# # Define a Reader Pool, load into memory
reader_pool = ReaderPool(app_config.config["reader"])

# # define the search index, load into memory
retriever_pool = RetrieverPool(app_config.config["retriever"])

# define the expander, load into memory
expander_pool = ExpanderPool(app_config.config["expander"])

handlers = Handler(reader_pool, retriever_pool, expander_pool)
# handlers = Handler(None, None)
api.include_router(handlers.router)





from neuralqa.utils import ConfigParser
import time
from fastapi import APIRouter
from typing import Optional
from neuralqa.server.routemodels import Document, Answer, Explanation
import logging

logger = logging.getLogger(__name__)


class Handler:
    def __init__(self, reader_pool, retriever_pool, expander_pool):
        router = APIRouter()
        self.router = router

        self.reader_pool = reader_pool
        self.retriever_pool = retriever_pool
        self.expander_pool = expander_pool

        @router.post("/answers")
        async def get_answers(params: Answer):

            """Generate an answer for the given search query.
            Performed as two stage process
            1.) Get sample passages from neighbourhood provided by matches by elastic search
            2.) Used BERT Model to identify exact answer spans

            Returns:
                [type] -- [description]
            """

            expanded_query = params.query

            answer_holder = []
            response = {}
            start_time = time.time()

            # switch to selected expander, perform expansion
            if params.expander != "none":
                self.expander_pool.selected_expander = params.expander
                if self.expander_pool.selected_expander:
                    expanded_query = self.expander_pool.expander.expand_query(
                        params.query)
                    params.query = params.query + " " + \
                        " ".join([term["token"]
                                  for term in expanded_query["terms"]])
                    expanded_query["query"] = params.query

            # switch to the selected model and retriever
            self.reader_pool.selected_model = params.reader
            self.retriever_pool.selected_retriever = params.retriever

            # answer question based on provided context
            if (params.retriever == "none" or self.retriever_pool.selected_retriever == None):
                answers = self.reader_pool.model.answer_question(
                    params.query, params.context, stride=params.tokenstride)
                for answer in answers:
                    answer["index"] = 0
                    answer_holder.append(answer)
            # answer question based on retrieved passages from elastic search
            else:

                num_fragments = 5
                query_results = self.retriever_pool.retriever.run_query(params.retriever, params.query,
                                                                        max_documents=params.max_documents, fragment_size=params.fragment_size,
                                                                        relsnip=params.relsnip, num_fragments=num_fragments, highlight_tags=False)
                # print(query_results)
                if (query_results["status"]):
                    # if relsnip is not enabled, read the entire document ... this is super slow
                    docs = query_results["highlights"] if params.relsnip else query_results["docs"]

                    for i, doc in enumerate(docs):
                        doc = doc.replace("\n", " ")
                        answers = self.reader_pool.model.answer_question(
                            params.query, doc, stride=params.tokenstride)
                        for answer in answers:
                            answer["index"] = i
                            answer_holder.append(answer)

                # sort answers by probability
                answer_holder = sorted(
                    answer_holder, key=lambda k: k['probability'], reverse=True)
            elapsed_time = time.time() - start_time
            response = {"answers": answer_holder,
                        "query": expanded_query, "took": elapsed_time}
            return response

        @ router.post("/documents")
        async def get_documents(params: Document):
            """Get a list of documents and highlights that match the given search query

            Returns:
                dictionary -- contains details on elastic search results.
            """

            num_fragments = 5
            query_results = {"docs": [], "highlights": []}

            self.retriever_pool.selected_retriever = params.retriever
            if self.retriever_pool.selected_retriever:
                query_results = self.retriever_pool.retriever.run_query(
                    params.retriever, params.query, max_documents=params.max_documents, fragment_size=params.fragment_size, relsnip=params.relsnip, num_fragments=num_fragments)
                # print(query_results)
                max_doc_size = 1200
                if not params.relsnip:
                    query_results["highlights"] = [
                        doc[:max_doc_size] + " .." for doc in query_results["docs"]]
            return query_results

        @ router.post("/explain")
        async def get_explanation(params: Explanation):
            """Return  an explanation for a given model

            Returns:
                [dictionary]: [explanation , query, question, ]
            """

            # TODO: Do we need to switch readers here?

            context = params.context.replace(
                "<em>", "").replace("</em>", "")

            gradients, token_words, token_types, answer_text = self.reader_pool.model.explain_model(
                params.query, context)

            explanation_result = {"gradients": gradients,
                                  "token_words": token_words,
                                  "token_types": token_types,
                                  "answer": answer_text
                                  }
            return explanation_result




from neuralqa.retriever import ElasticSearchRetriever
import logging

logger = logging.getLogger(__name__)


class RetrieverPool():
    def __init__(self, retrievers):

        self.retriever_pool = {}
        for retriever in retrievers["options"]:
            if (retriever["value"] in self.retriever_pool):
                raise ValueError(
                    "Duplicate retriever value : {} ".format(retriever["value"]))

            if (retriever["type"] == "elasticsearch"):
                self.retriever_pool[retriever["value"]] = ElasticSearchRetriever(
                    host=retriever["host"], port=retriever["port"], body_field=retriever["fields"]["body_field"])
            if (retriever["type"] == "solr"):
                logger.info("We do not yet support Solr retrievers")
        self.selected_retriever = retrievers["selected"]

    @property
    def retriever(self):
        return self.retriever_pool[self.selected_retriever]

    @property
    def selected_retriever(self):
        return self._selected_retriever

    @selected_retriever.setter
    def selected_retriever(self, selected_retriever):
        if (selected_retriever in self.retriever_pool):
            self._selected_retriever = selected_retriever
        else:
            if (len(self.retriever_pool) > 0):
                default_retriever = next(iter(self.retriever_pool))
                logger.info(
                    ">> Retriever you are attempting to use (%s) does not exist in retriever pool. Using the following default retriever instead %s ", selected_retriever, default_retriever)
                self._selected_retriever = default_retriever
            else:
                logger.info(
                    ">> No retriever has been specified in config.yaml.")
                self._selected_retriever = None





from pydantic import BaseModel
from typing import Optional


class Document(BaseModel):

    max_documents: Optional[int] = 5
    query: str = "what is a fourth amendment right violation? "
    fragment_size: int = 250
    retriever: Optional[str] = None
    relsnip: Optional[bool] = True


class Answer(BaseModel):

    max_documents: Optional[int] = 5
    query: str = "what is a fourth amendment right violation? "
    fragment_size: int = 250
    tokenstride: int = 50
    context: Optional[str] = "The fourth amendment kind of protects the rights of citizens .. such that they dont get searched"
    reader: str = None
    relsnip: bool = True
    expander: Optional[str] = None
    retriever: Optional[str] = "manual"


class Explanation(BaseModel):
    query: str = "what is a fourth amendment right violation? "
    context: str = "The fourth amendment kind of protects the rights of citizens .. such that they dont get searched"



import os
from importlib.machinery import SourceFileLoader
from setuptools import setup, find_packages


version = SourceFileLoader('neuralqa.version', os.path.join(
    'neuralqa', 'version.py')).load_module().VERSION


def package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


ui_files = package_files("neuralqa/server/ui/build")
yaml_file = ["config_default.yaml"]
setup(
    name='neuralqa',
    packages=find_packages(exclude=['tests', 'tests.*']),
    package_data={"neuralqa": ui_files + yaml_file},
    version=version,
    license='MIT',
    description='NeuralQA: Question Answering on Large Datasets',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    author='Victor Dibia',
    url='https://github.com/victordibia/neuralqa',
    python_requires='>=3.5',
    # download_url='https://github.com/victordibia/neuralqa/archive/v0.0.2.tar.gz',
    keywords=['NLP', 'Question Answering', 'Machine Learning'],
    install_requires=[
        'fastapi',
        'aiofiles',
        'uvicorn',
        'numpy',
        'tensorflow>=2.1.0',
        'torch',
        'torchvision',
        'transformers',
        'elasticsearch>=7.7.1',
        'pyyaml>=3.13',
        'spacy'
    ],
    extras_require={
        'test': ['pytest']
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],
    entry_points={
        "console_scripts": [
            "neuralqa=neuralqa.cli:cli",
        ]
    }
)



import click
from neuralqa.server import launch_server
from neuralqa.utils import cli_args
from neuralqa.utils import import_sample_data, ConfigParser
import os
from neuralqa.retriever import RetrieverPool
import logging


@click.group()
@click.version_option()
def cli():
    pass


# @cli.command()
# @cli_args.HOST
# @cli_args.PORT
# @cli_args.WORKERS
# @cli_args.CONFIG_PATH
# def test(host, port, workers, config_path):
#     import_sample_data()


@cli.command()
@cli_args.MAX_DOCS
def load(max_docs):
    """This command loads sample data into a local elastic search index."""
    import_sample_data(max_docs=max_docs)


@cli.command()
@cli_args.HOST
@cli_args.PORT
@cli_args.WORKERS
@cli_args.CONFIG_PATH
def ui(host, port, workers, config_path):
    """This command launches the web interface for NeuralQA."""
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    logging.getLogger(__name__).setLevel(logging.INFO)
    if (config_path):
        os.environ["NEURALQA_CONFIG_PATH"] = config_path
    launch_server(host, port, workers)


if __name__ == '__main__':
    cli()



import logging
from neuralqa.version import VERSION as __version__
from neuralqa.reader import BERTReader
from neuralqa.utils import import_sample_data


logging.getLogger("transformers").setLevel(logging.ERROR) 
logging.getLogger("tensorflow").setLevel(logging.ERROR) 
logging.getLogger("elasticsearch").setLevel(logging.CRITICAL)

__all__ = ["BERTReader", "import_sample_data"]




import uvicorn
import os


def launch_server(host="127.0.0.1", port=5000, workers=1, reload=False):
    uvicorn.run("neuralqa.server.serve:app", host=host, port=port, workers=workers,
                log_level="info", reload=reload)


if __name__ == "__main__":
    launch_server()
